from collections import defaultdict as deft
from doctest import testmod
from typing import Any, Dict, List, Union

from unidecode import unidecode


class Trie:

    def __init__(
        self,
        case_sensitive: bool = False,
        decode_ascii: bool = True
    ) -> None:
        """
        Constructor for an instance of a Trie.

        Parameters
        ----------
        case_sensitive: bool
            If set to true, look-up on the Trie is case-sensitive. It is
            insensitive otherwise (and by default).

        decode_ascii: bool
            If set to false, non-ASCII strings are first decoded into ASCII strings
            are then added to the Trie. The Trie will thus return matches for the
            ASCII strings, not the original ones. No mapping between both encodings
            is kept inside the Trie and it should be implemented if needed.


        Examples
        --------
        >>> trie = Trie()
        >>> trie += [
        ...    "orca", "Orco", "orco", "oro",
        ...    "orwelliano", "oráculo", "oración",
        ... ]
        >>> assert trie('a') == []
        >>> assert trie.search("a") == []
        >>> assert trie('or') == []
        >>> assert trie.search("or") == [
        ...   'orca', 'orco', 'oro',
        ...   'orwelliano', 'oraculo', 'oracion'
        ... ]
        >>> assert trie('ora') == []
        >>> assert trie.search("ora") == ['oraculo', 'oracion']
        >>> assert trie('orco') == ['orco']
        >>> assert trie.search("orco") == ['orco']

        >>> trie = Trie(case_sensitive=True)
        >>> trie += [
        ...    "orca", "Orco", "orco", "oro", "Orwell",
        ...    "orwelliano", "oráculo", "oración"
        ... ]
        >>> assert trie.search('Or') == ['Orco', 'Orwell']
        >>> assert trie('ora') == []
        >>> assert trie.search("ora") == ['oraculo', 'oracion']
        >>> assert trie('Orco') == ['Orco']
        >>> assert trie.search("orco") == ['orco']

        >>> trie = Trie(decode_ascii=False)
        >>> trie += [
        ...    "orca", "Orco", "orco", "oro",
        ...    "orwelliano", "oráculo", "oración",
        ... ]
        >>> assert trie('ora') == []
        >>> assert trie.search("ora") == ['oración']
        >>> assert trie("oracion") == []
        >>> assert trie.search("oracion") == []
        >>> assert trie("oración") == ['oración']
        >>> assert trie.search("oració") == ['oración']

        """
        self.case_sensitive = case_sensitive
        self.decode_ascii = decode_ascii
        self._tree = dict([])

    def __iadd__(self, words: List[str]) -> Any:
        for word in words:
            self.add(word)
        return self

    def __preprocess_word(self, word: str) -> str:
        word = word.lower() if not self.case_sensitive else word
        if self.decode_ascii:
            word = unidecode(word)
        return word

    def add(self, word: str) -> None:
        word = self.__preprocess_word(word)
        parts = list(word)
        tree = self._tree
        while parts:
            part = parts.pop(0)
            if part not in tree:
                tree[part] = dict([])
            tree = tree[part]
            if not parts:
                tree["#"] = True

    def __contains__(self, word: str) -> List[str]:
        return self(w)

    def search(self, word: str) -> List[str]:
        word = self.__preprocess_word(word)
        candidates = self.__lookup(
            word, self._tree, [], list(word), []
        )
        return candidates

    def __call__(self, word: str) -> List[str]:
        candidates = self.search(word)
        return [candidate for candidate in candidates if candidate == word]

    def __lookup(
        self,
        word: str,
        tree: Dict[str, Union[str, Dict]],
        history: List[str],
        remainder: List[str],
        results: List[str],
    ) -> List[str]:
        part = remainder.pop(0)
        if part not in tree:
            return results
        else:
            tree = tree[part]
            if remainder:
                return self.__lookup(
                    word, tree, history + [part], remainder, results
                )
            else:
                return results + self.__pull_all_children(history + [part], tree)

    def __pull_all_children(
        self,
        history: List[str],
        tree: List[str]
    ) -> List[str]:
        _children = []
        for key, val in tree.items():
            if key == "#":
                _children.append("".join(history))
                continue
            _child = [ch for ch in history] + [key]
            if val.keys():
                _children += self.__pull_all_children(_child, val)
            else:
                _children.append(_child)
        return _children



def test_speed():

    import math
    import random
    import time

    letters = list('qwertyuiopasdfghjklzxcvbnm')

    def random_word():
        letter_indexes = [
            random.randrange(len(letters))
            for _ in range(random.randrange(3, 15))
        ]
        return ''.join([letters[i] for i in letter_indexes])

    vocab = [random_word() for _ in range(10000)]
    test_set = [word for word in vocab]
    for pool_oversample, times_oversampled in [
        (random.sample(vocab, 100), 1000),
        (random.sample(vocab, 1000), 100),
        (random.sample(vocab, 10000), 10)
    ]:
        for word in pool_oversample:
            test_set += [word] * times_oversampled

    n_eval = 10000
    trie = Trie()
    trie += test_set
    start = time.time()
    trie.add("Arda")
    for w in random.sample(vocab + ["Ard"], n_eval):
        results = trie.search(w)
        if w in "Arda":
            assert True
    runtime = time.time() - start
    throughput_second = n_eval * (1 / runtime)

    assert runtime < 0.05
    assert round(math.log(throughput_second, 10)) >= 5



if __name__ == '__main__':
    testmod()
    test_speed()


from collections import OrderedDict
import json
from pickle import (
    dump as pdump,
    load as pload
)

from doctest import testmod
from itertools import chain
from typing import Any, Dict, List

from .trie import Trie


def batched(iterable: List[Any], batch_size: int) -> List[List[Any]]:
    """
    Partitions an input collection `iterable` into chunks of size `batch_size`.
    The number of chunks is unknown at the time of calling is determined by
    the length of `iterable`.

    Parameters
    ----------
    iterable:   List[Any]

    batch_size: int

    Returns
    -------
    List[List[Any]]

    Examples
    --------
    >>> iterable = [1, 2, 3, 4, 5, 6, 7, 8]
    >>> chunks = batched(iterable, batch_size=2)
    >>> assert len(chunks) == 4
    >>> assert chunks[0] == [1, 2]
    >>> assert chunks[-1] == [7, 8]

    >>> iterable = [1, 2, 3, 4, 5, 6, 7, 8]
    >>> chunks = batched(iterable, batch_size=12)
    >>> assert len(chunks) == 1
    >>> assert chunks[0] == iterable

    >>> iterable = [1, 2, 3]
    >>> chunks = batched(iterable, batch_size=2)
    >>> assert chunks == [
    ...    [1, 2],
    ...    [3]
    ... ]

    """
    idxs = list(range(len(iterable)))
    ii = [i for i in idxs[::batch_size]]
    return [iterable[i:i + batch_size] for i in ii]


def flatten_loop(lists):
    flattened = []
    for l in lists:
        flattened.extend(l)
    return flattened


def flatten_func(lists):
    return list(chain(*lists))


def flatten(lists: List[List[Any]]) -> List[Any]:
    """
    Given a collection of lists, concatenates all elements into a single list.

    More formally, given a collection holding `n` iterables with `m` elements
    each, this function will return a single list holding all `n * m` elements.

    Parameters
    ----------
    List[List[Any]]

    Returns
    -------
    List[Any]

    Examples
    --------
    >>> example = [[1, 2, 3], [1], [2, 4, 6], [3, 6, 9], [7, 13]]
    >>> len_example = sum(len(l) for l in example)

    >>> assert len_example == len(flatten(example))
    >>> assert len_example == len(flatten_func(example))
    >>> assert len_example == len(flatten_loop(example))

    >>> assert flatten(example) == flatten_func(example)

    >>> assert flatten(example) == flatten_loop(example)
    """
    return [e for l in lists for e in l]


def to_txt(string: str, path: str) -> None:
    """
    Function that expects two string parameters as arguments and writes the
    first string as the content of a file at the location denoted by the second
    string (which is assumed to denote a POSIX path).

    Parameters
    ----------
    string: str
        Some text data to write to disk.

    path: str
        The location where the input text data must be stored, as a POSIX path.

    Returns
    -------
    Nothing, writes the value stored in input variable `string` to the disk
    location denoted by `path`.

    Examples
    --------
    >>> import os
    >>> test_path = "./test_path.txt"

    >>> assert not os.path.exists(test_path)
    >>> to_txt("test raw text.", test_path)
    >>> assert os.path.exists(test_path)
    >>> assert os.path.isfile(test_path)
    >>> assert from_txt(test_path) == "test raw text."

    >>> os.remove(test_path)
    """
    with open(path, 'w') as wrt:
        wrt.write(string)


def from_txt(path: str) -> str:
    """
    Function that can be directed to a local raw text file by its POSIX path
    and returns the content of that file as a string.

    Parameters
    ----------
    path: str
        The location where the input text data must be stored, as a POSIX path.

    Returns
    -------
    str: the raw-text content read from the disk location denoted by the
    argument of parameter `path`.

    Examples
    --------
    >>> import os
    >>> test_path = "./test_path.txt"

    >>> assert not os.path.exists(test_path)
    >>> to_txt("test raw text.", test_path)
    >>> assert os.path.exists(test_path)
    >>> assert os.path.isfile(test_path)
    >>> assert from_txt(test_path) == "test raw text."

    >>> os.remove(test_path)
    """
    with open(path, 'r') as rd:
        return rd.read().strip()


def from_json(path: str) -> Dict[Any, Any]:
    """
    Function that can be directed to a local raw text file by its POSIX path
    and returns the content of that file as a Python dictionary.

    Parameters
    ----------
    path: str
        The location where the input text data must be stored, as a POSIX path.

    Returns
    -------
    str: the dictionary content read from the disk location denoted by the
    argument of parameter `path`.

    Examples
    --------
    >>> import os
    >>> path_json = "json_dump.json"
    >>> assert not os.path.exists(path_json)

    >>> data = {1: "one", 2: "two", 3: "three"}
    >>> to_json(data, path_json)
    >>> assert os.path.exists(path_json)

    >>> keys = list(data.keys())
    >>> for key in keys:
    ...    del data[key]
    >>> assert len(data) == 0

    >>> data.update(from_json(path_json))
    >>> assert len(data) == 3
    >>> assert data == OrderedDict({"1": "one", "2": "two", "3": "three"})

    >>> os.remove(path_json)
    """
    with open(path, 'r') as rd:
        data = json.load(rd, object_pairs_hook=OrderedDict)
    return data


def to_json(dict_: Dict[Any, Any], path: str, indentation: int = 4) -> None:
    """
    Function that expects two parameters as arguments, a Python dictionary and
    a string, and writes the former as the content of a file at the location
    denoted by the latter (which is assumed to denote a POSIX path).

    Parameters
    ----------
    dict_: Any
        A Python dictionary (associative array) whose contents we want
        serialized to disk. The contents must be JSON-dumpable, e.g. no keys
        or values in the dictionary should contain binaries. Otherwise,
        consider pickling the object with `to_pickle`.

    path: str
        The location where the input text data must be stored, as a POSIX path.

    indentation: int
        An integer denoting the indentation to use for every level of nested
        dictionaries stored in input object `dict_`. A dictionary consisting
        of a keys and values will be serialized with an indentation equal to
        `indentation x 1` whitespace characters. If any of those values itself
        contains another dictionary, the values of the latter will be
        serialized with an indentation level equal to `indentation x 2`, and
        so on.

    Returns
    -------
    Nothing, writes the value stored in input variable `payload` to the disk
    location denoted by `path`.

    Examples
    --------
    >>> import os
    >>> path_json = "json_dump.json"
    >>> assert not os.path.exists(path_json)

    >>> data = {1: "one", 2: "two", 3: "three"}
    >>> to_json(data, path_json)
    >>> assert os.path.exists(path_json)

    >>> keys = list(data.keys())
    >>> for key in keys:
    ...    del data[key]
    >>> assert len(data) == 0

    >>> data.update(from_json(path_json))
    >>> assert len(data) == 3
    >>> assert data == OrderedDict({"1": "one", "2": "two", "3": "three"})

    >>> os.remove(path_json)
    """
    with open(path, 'w') as wrt:
      json.dump(dict_, wrt, indent=indentation)



def to_pickle(data: Any, path: str) -> None:
    """
    Parameters
    ----------

    Returns
    -------

    Examples
    --------
    >>> import os
    >>> path_pickle = "pickle_file.p"
    >>> assert not os.path.exists(path_pickle)

    >>> data = {1: "one", 2: "two", 3: "three"}
    >>> to_pickle(data, path_pickle)
    >>> assert os.path.exists(path_pickle)

    >>> keys = list(data.keys())
    >>> for key in keys:
    ...    del data[key]
    >>> assert len(data) == 0

    >>> try:
    ...   from_json(path_pickle)
    ... except Exception:
    ...   assert True

    >>> try:
    ...   from_txt(path_pickle)
    ... except Exception:
    ...   assert True

    >>> data.update(from_pickle(path_pickle))
    >>> assert len(data) == 3
    >>> assert data == {1: "one", 2: "two", 3: "three"}

    >>> os.remove(path_pickle)
    """
    with open(path, "wb") as wrt:
        pdump(data, wrt)


def from_pickle(path: str):
    """
    Parameters
    ----------

    Returns
    -------

    Examples
    --------
    >>> import os
    >>> path_pickle = "pickle_file.p"
    >>> assert not os.path.exists(path_pickle)

    >>> data = {1: "one", 2: "two", 3: "three"}
    >>> to_pickle(data, path_pickle)
    >>> assert os.path.exists(path_pickle)

    >>> keys = list(data.keys())
    >>> for key in keys:
    ...    del data[key]
    >>> assert len(data) == 0

    >>> try:
    ...   from_json(path_pickle)
    ... except Exception:
    ...   assert True

    >>> try:
    ...   from_txt(path_pickle)
    ... except Exception:
    ...   assert True

    >>> data.update(from_pickle(path_pickle))
    >>> assert len(data) == 3
    >>> assert data == {1: "one", 2: "two", 3: "three"}

    >>> os.remove(path_pickle)
    """
    with open(path, "rb") as rd:
        data = pload(rd)
    return data




if __name__ == '__main__':

    testmod()

